import logging
import os
from typing import Dict

from dotenv import load_dotenv
from flask import current_app, Flask, jsonify
from flask_migrate import Migrate
from marshmallow import ValidationError
from raven import Client
from raven.contrib.flask import Sentry

from stumpf.auth import auth_bp, before_request_auth, login_manager
from stumpf.index import bp as index_bp
from stumpf.mock_auth import mock_auth_bp
from stumpf.models import db
from stumpf.views import (annotation_app_bp, artifact_bp, assignment_bp, task_bp, token_bp, user_bp,
                          video_bp, webhook_bp)


def handle_validation_error(e):
    return jsonify(e.messages), 400


def handle_general_error(e):
    current_app.logger.exception(e)
    return jsonify({'error': 'Something went wrong.'}), 500


def create_app(config: Dict) -> Flask:
    load_dotenv(os.getenv('DOTENV_PATH'))

    app = Flask(__name__)
    app.config.update(config)
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['SQLALCHEMY_DATABASE_URI'] = os.getenv('STUMPF_POSTGRES_URI')
    app.config['SECRET_KEY'] = os.getenv('SECRET_KEY')
    app.config['GOOGLE_OAUTH_CLIENT_ID'] = os.getenv('GOOGLE_OAUTH_CLIENT_ID')
    app.config['GOOGLE_OAUTH_CLIENT_SECRET'] = os.getenv('GOOGLE_OAUTH_CLIENT_SECRET')

    app.config['MTURK_SANDBOX_ENDPOINT_URL'] = os.getenv('MTURK_SANDBOX_ENDPOINT_URL')

    if app.config['ENV'] == 'development':
        app.logger.info(
            'Ignoring MTURK_PRODUCTION_ENDPOINT_URL, real HITs can only be launched in production.')
    else:
        app.config['MTURK_PRODUCTION_ENDPOINT_URL'] = os.getenv('MTURK_PRODUCTION_ENDPOINT_URL')

    app.config['MTURK_SNS_ARN'] = os.getenv('MTURK_SNS_ARN')

    if os.getenv('SENTRY_DSN', False):
        client = Client(os.getenv('SENTRY_DSN'), environment=app.config['ENV'],
                        processors=('raven.processors.SanitizePasswordsProcessor',
                                    'stumpf.sentry.SanitizeProcessor'))
        app.sentry = Sentry(app, client, logging=True, level=logging.WARN)

    if app.config['ENV'] == 'development':
        app.config['SQLALCHEMY_ECHO'] = True

    db.init_app(app)
    Migrate(app, db)
    login_manager.init_app(app)

    app.register_blueprint(index_bp, url_prefix='/')
    app.register_blueprint(webhook_bp, url_prefix='/webhooks')

    app.register_blueprint(annotation_app_bp, url_prefix='/api/v1')
    app.register_blueprint(artifact_bp, url_prefix='/api/v1')
    app.register_blueprint(assignment_bp, url_prefix='/api/v1')
    app.register_blueprint(task_bp, url_prefix='/api/v1')
    app.register_blueprint(token_bp, url_prefix='/api/v1')
    app.register_blueprint(user_bp, url_prefix='/api/v1')
    app.register_blueprint(video_bp, url_prefix='/api/v1')

    # The default behavior in development mode will be to mock out google's
    # oauth service.  If a client secret key is set, the real oauth provider
    # will be used.
    if app.config['ENV'] == 'development' and \
            not app.config['GOOGLE_OAUTH_CLIENT_SECRET']:
        app.logger.info('Using mocked oauth service')
        app.register_blueprint(mock_auth_bp, url_prefix='/auth')
    else:
        app.logger.info('Using real oauth service')
        app.register_blueprint(auth_bp, url_prefix='/auth')

    app.before_request(before_request_auth)

    # Only mask all errors/exceptions when in production
    if app.config['ENV'] == 'production':
        app.register_error_handler(500, handle_general_error)

    app.register_error_handler(ValidationError, handle_validation_error)

    return app
