from flask import abort, redirect, request
from flask_dance.consumer import oauth_authorized
from flask_dance.consumer.backend.sqla import SQLAlchemyBackend
from flask_dance.contrib.google import make_google_blueprint
from flask_login import current_user, login_user
from flask_login import LoginManager
from sqlalchemy.orm.exc import NoResultFound

from stumpf.decorators import is_public
from stumpf.models import db, OAuth, Token, User

auth_bp = make_google_blueprint(backend=SQLAlchemyBackend(OAuth, db.session, user=current_user),
                                reprompt_consent=True, offline=True, scope=['email'])
login_manager = LoginManager()


def before_request_auth(*args, **kwargs):
    if is_public():
        return

    user = Token.user_from_token(request.headers.get('X-Stumpf-Token'))
    if user is None:
        login_manager.unauthorized()
    login_user(user)


@login_manager.user_loader
def load_user(user_id):
    return User.query.get(user_id)


# create/login local user on successful OAuth login
def google_logged_in(blueprint, token):
    if not token:
        abort(403)

    resp = blueprint.session.get('/oauth2/v2/userinfo')
    if not resp.ok:
        abort(403)

    google_info = resp.json()

    if google_info['hd'] != 'kitware.com':
        abort(403)

    oauth_query = OAuth.query.filter_by(
        provider=blueprint.name,
        provider_user_id=google_info['id']
    )

    try:
        oauth = oauth_query.one()
    except NoResultFound:
        oauth = OAuth(
            provider=blueprint.name,
            provider_user_id=google_info['id'],
            token=token
        )

    if oauth.user:
        login_user(oauth.user)
    else:
        try:
            user = User(email=google_info['email'])
            oauth.user = user
            db.session.add_all([user, oauth])
            db.session.commit()
            login_user(user)
        except Exception:
            db.session.rollback()
            raise

    stumpf_token = Token(user_id=current_user.get_id())
    db.session.add(stumpf_token)
    db.session.commit()

    response = redirect(request.args.get('next', '/'))
    response.set_cookie('stumpf-token', stumpf_token.token, expires=stumpf_token.expires)
    return response


@oauth_authorized.connect_via(auth_bp)
def google_logged_in_decorated(*args, **kwargs):
    return google_logged_in(*args, **kwargs)
