from datetime import datetime, timedelta
import secrets
from typing import Optional
import uuid

from flask_dance.consumer.backend.sqla import OAuthConsumerMixin
from flask_sqlalchemy import SQLAlchemy
from marshmallow import fields, post_load, Schema, validate
from sqlalchemy import MetaData
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.schema import UniqueConstraint


# This is to avoid having to manually name all constraints
# See: http://alembic.zzzcomputing.com/en/latest/naming.html
metadata = MetaData(naming_convention={
    'pk': 'pk_%(table_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'ix': 'ix_%(table_name)s_%(column_0_name)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
})
db = SQLAlchemy(metadata=metadata)


class BaseSchema(Schema):
    __model__ = None

    id = fields.UUID(dump_only=True)
    created = fields.DateTime(dump_only=True)

    @post_load
    def make_object(self, data):
        return self.__model__(**data)


class User(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    email = db.Column(db.String, unique=True, nullable=False)

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return True

    @property
    def is_anonymous(self):
        return not self.is_authenticated

    def get_id(self):
        return self.id


class UserSchema(BaseSchema):
    __model__ = User

    email = fields.Email(required=True)


class OAuth(OAuthConsumerMixin, db.Model):
    provider_user_id = db.Column(db.String(256), unique=True)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(User.id))
    user = db.relationship(User)


def now_plus(**kwargs):
    def f():
        return datetime.utcnow() + timedelta(**kwargs)
    return f


class Token(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    user_id = db.Column(db.ForeignKey(User.id, name='token_user_id_fkey'), nullable=False)
    user = db.relationship(User)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False, default=now_plus(days=7))
    token = db.Column(db.String, index=True, unique=True, nullable=False, default=secrets.token_hex)

    @classmethod
    def user_from_token(cls, token: str) -> Optional[User]:
        t = cls.query.filter_by(token=token).filter(cls.expires > datetime.utcnow()).first()

        if t:
            return t.user


class TokenSchema(BaseSchema):
    __model__ = Token

    user_id = fields.UUID(load_only=True)
    user = fields.Nested('UserSchema', dump_only=True)
    expires = fields.DateTime()
    token = fields.Str(dump_only=True)


class Video(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    path = db.Column(db.String, unique=True, nullable=False)
    name = db.Column(db.String, unique=True, nullable=False)
    frame_rate = db.Column(db.Float, nullable=False)
    duration = db.Column(db.Float, nullable=False)
    width = db.Column(db.Integer, nullable=False)
    height = db.Column(db.Integer, nullable=False)

    derived_video_id = db.Column(db.ForeignKey('video.id'), nullable=True)
    derived_video = db.relationship('Video', remote_side=[id],
                                    backref=db.backref('derivations'))

    sortable_fields = {
        'created': created,
        'path': path,
        'name': name,
        'duration': duration
    }
    default_sort = 'name'
    default_desc = False


class VideoSchema(BaseSchema):
    __model__ = Video

    path = fields.Str(required=True)
    name = fields.Str(required=True)
    frame_rate = fields.Float(required=True)
    duration = fields.Float(required=True)
    width = fields.Integer(required=True)
    height = fields.Integer(required=True)

    derived_video_id = fields.UUID(load_only=True)
    derived_video = fields.Nested('VideoSchema', dump_only=True)


class AnnotationApp(db.Model):
    __table_args__ = (UniqueConstraint('repo_url', 'hash', 'name', 'url'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    repo_url = db.Column(db.String, nullable=False)
    hash = db.Column(db.String, nullable=False)
    name = db.Column(db.String, nullable=False)
    url = db.Column(db.String, nullable=False)

    sortable_fields = {
        'created': created,
        'name': name
    }
    default_sort = 'created'
    default_desc = True


class AnnotationAppSchema(BaseSchema):
    __model__ = AnnotationApp

    repo_url = fields.Url(required=True)
    url = fields.Url(required=True)
    hash = fields.Str(required=True)
    name = fields.Str(required=True)


class Task(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('user.id'), nullable=False)
    user = db.relationship('User', backref=db.backref('tasks', lazy=True))
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False,
                         index=True)
    video = db.relationship('Video', backref=db.backref('tasks', lazy=True))
    annotation_app_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('annotation_app.id'),
                                  nullable=False)
    annotation_app = db.relationship('AnnotationApp', backref=db.backref('tasks', lazy=True))
    hit_id = db.Column(db.String, nullable=True)
    type_ = db.Column(pg.ENUM('HIT', 'KWIT', name='hit_kwit_enum'), nullable=False)
    title = db.Column(db.String, nullable=False)
    description = db.Column(db.String, nullable=True)
    keywords = db.Column(db.String, nullable=True)
    status = db.Column(pg.ENUM('Assignable', 'Unassignable', 'Reviewable', 'Reviewing', 'Disposed',
                               name='task_status_enum'), nullable=False, default='Assignable')
    review_status = db.Column(pg.ENUM('NotReviewed', 'MarkedForReview', 'ReviewedAppropriate',
                                      'ReviewedInappropriate', name='task_review_status_enum'),
                              nullable=False, default='NotReviewed')
    max_assignments = db.Column(db.Integer, nullable=False)
    reward = db.Column(db.Integer, nullable=True)  # cents
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})

    sortable_fields = {
        'created': created,
        'title': title,
        'type_': type_
    }
    default_sort = 'created'
    default_desc = True


class TaskSchema(BaseSchema):
    __model__ = Task

    annotation_app = fields.Nested('AnnotationAppSchema', dump_only=True)
    title = fields.Str(required=True)
    type_ = fields.Str(required=True)
    video_id = fields.UUID(required=True, load_only=True)
    video = fields.Nested('VideoSchema', dump_only=True)
    hit_id = fields.Str()
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    status = fields.Str()
    review_status = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=300))


class CreateHitSchema(Schema):
    sandbox = fields.Bool(default=True)
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    assignment_duration_in_seconds = fields.Int(required=True)
    lifetime_in_seconds = fields.Int(required=True)
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=300))
    args = fields.Dict(keys=fields.Str())


class Assignment(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False)
    assignment_id = db.Column(db.String, nullable=True)
    worker_id = db.Column(db.String, nullable=True)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), nullable=False)
    accepted_time = db.Column(db.DateTime, nullable=True)
    abandoned_time = db.Column(db.DateTime, nullable=True)
    returned_time = db.Column(db.DateTime, nullable=True)
    submitted_time = db.Column(db.DateTime, nullable=True)
    approved_time = db.Column(db.DateTime, nullable=True)
    rejected_time = db.Column(db.DateTime, nullable=True)

    task = db.relationship('Task', backref=db.backref('assignments', lazy=True))

    sortable_fields = {
        'created': created
    }
    default_sort = 'created'
    default_desc = True


class AssignmentSchema(BaseSchema):
    __model__ = Assignment

    task = fields.Nested('TaskSchema', dump_only=True)
    assignment_id = fields.Str()
    worker_id = fields.Str()
    status = fields.Str()
    accepted_time = fields.DateTime()
    abandoned_time = fields.DateTime()
    returned_time = fields.DateTime()
    submitted_time = fields.DateTime()
    approved_time = fields.DateTime()
    rejected_time = fields.DateTime()


class Artifact(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('assignment.id'), nullable=False)
    path = db.Column(db.String, unique=True, nullable=False)

    assignment = db.relationship('Assignment', backref=db.backref('artifacts', lazy=True))

    sortable_fields = {
        'created': created,
        'path': path
    }
    default_sort = 'created'
    default_desc = True


class ArtifactSchema(BaseSchema):
    __model__ = Artifact

    assignment_id = fields.Str(required=True)
    path = fields.Str(required=True)
