from base64 import b64decode
from datetime import datetime
from typing import Dict

from flask import current_app
from jinja2 import Template
from M2Crypto import X509
import requests

from stumpf.models import Assignment, db, Task


external_question_tpl = Template("""<?xml version="1.0"?><ExternalQuestion xmlns="http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2006-07-14/ExternalQuestion.xsd"><ExternalURL>{{ app_url }}?{{ query_string|escape }}</ExternalURL><FrameHeight>800</FrameHeight></ExternalQuestion>""")  # noqa


def question_xml(**kwargs):
    return external_question_tpl.render(**kwargs)


def generate_message(data: Dict) -> str:
    """
    Takes an SNS data value and generates the signable message.

    This is according to the specification detailed here:
    https://docs.aws.amazon.com/sns/latest/dg/SendMessageToHttp.verify.signature.html
    """
    message_keys = {'Notification': ['Message',
                                     'MessageId',
                                     'Subject',
                                     'Timestamp',
                                     'TopicArn',
                                     'Type'],
                    'SubscriptionConfirmation': ['Message',
                                                 'MessageId',
                                                 'SubscribeURL',
                                                 'Timestamp',
                                                 'Token',
                                                 'TopicArn',
                                                 'Type'],
                    'UnsubscribeConfirmation': ['Message',
                                                'MessageId',
                                                'SubscribeURL',
                                                'Timestamp',
                                                'Token',
                                                'TopicArn',
                                                'Type']}

    keys = message_keys[data['Type']]
    message = ''

    for key in keys:
        message += f'{key}\n{data[key]}\n'

    return message


def verify_signature(data: Dict) -> bool:
    message = generate_message(data)
    signature = b64decode(data['Signature'])

    cert = X509.load_cert_string(requests.get(data['SigningCertURL']).text)
    pubkey = cert.get_pubkey()
    pubkey.reset_context(md='sha1')
    pubkey.verify_init()
    pubkey.verify_update(message.encode())

    return True if pubkey.verify_final(signature) == 1 else False


def handle_mturk_assignment_event(event: Dict):
    event['EventTimestamp'] = datetime.strptime(event['EventTimestamp'], '%Y-%m-%dT%H:%M:%SZ')

    # get or create the assignment
    if event['EventType'] == 'AssignmentAccepted':
        task = Task.query.filter_by(hit_id=event['HITId']).first()

        if not task:
            current_app.logger.warning(f'Received event with unknown task, event: {event}')
            return

        assignment = Assignment(task=task, assignment_id=event['AssignmentId'],
                                worker_id=event['WorkerId'], status='Accepted',
                                accepted_time=event['EventTimestamp'])
    else:
        assignment = Assignment.query.filter_by(assignment_id=event['AssignmentId']).first()

        if not assignment:
            current_app.logger.warning(f'Received event for unknown assignment, event: {event}')
            return

    if event['EventType'] == 'AssignmentAbandoned':
        assignment.status = 'Abandoned'
        assignment.abandoned_time = event['EventTimestamp']
    elif event['EventType'] == 'AssignmentReturned':
        assignment.status = 'Returned'
        assignment.returned_time = event['EventTimestamp']
    elif event['EventType'] == 'AssignmentSubmitted':
        assignment.status = 'Submitted'
        assignment.submitted_time = event['EventTimestamp']
    elif event['EventType'] == 'AssignmentRejected':
        assignment.status = 'Rejected'
        assignment.rejected_time = event['EventTimestamp']
    elif event['EventType'] == 'AssignmentApproved':
        assignment.status = 'Approved'
        assignment.approved_time = event['EventTimestamp']
    elif event['EventType'] != 'AssignmentAccepted':  # catch all
        current_app.logger.warning('Received unknown event type', event)
        return

    try:
        db.session.add(assignment)
        db.session.commit()
    except Exception as e:
        db.session.rollback()
        current_app.logger.warning(f'Failed to create/update assignment, event: {event}')
        raise


def handle_mturk_hit_event(event: Dict):
    task = Task.query.filter_by(hit_id=event['HITId']).first()

    if not task:
        current_app.logger.warning('Received event for unknown task', event)
        return
    elif event['EventType'] == 'HITCreated':
        return  # sanity check to ensure the HIT exists

    if event['EventType'] == 'HITReviewable':
        task.status = 'Reviewable'
    elif event['EventType'] == 'HITExpired':
        task.status = 'Unassignable'
    elif event['EventType'] == 'HITDisposed':
        task.status = 'Disposed'
    else:
        current_app.logger.warning('Received unknown event type', event)
        return

    try:
        db.session.add(task)
        db.session.commit()
    except Exception as e:
        current_app.logger.warning('Failed to update task', event)
        raise
