from urllib.parse import urlencode

import boto3
from flask import Blueprint, current_app, request
from flask_login import current_user

from stumpf.models import (AnnotationApp, Assignment, AssignmentSchema, CreateHitSchema, db, Task,
                           TaskSchema, Video)
from stumpf.mturk_utils import question_xml
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query


task_bp = Blueprint('task', __name__)

assignment_schema = AssignmentSchema()
task_schema = TaskSchema()
create_hit_schema = CreateHitSchema()


@task_bp.route('/task/<uuid:video_id>/<uuid:app_id>/hit', methods=['POST'])
def create_task_hit(video_id, app_id):
    try:
        video = Video.query.get_or_404(video_id)
        annotation_app = AnnotationApp.query.get_or_404(app_id)
        hit_args = create_hit_schema.load(request.json)

        if hit_args['sandbox']:
            mturk = boto3.client('mturk',
                                 endpoint_url=current_app.config['MTURK_SANDBOX_ENDPOINT_URL'])
        else:
            mturk = boto3.client('mturk',
                                 endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])

        hit_type_r = mturk.create_hit_type(
            AssignmentDurationInSeconds=hit_args['assignment_duration_in_seconds'],
            Reward=str(hit_args['reward'] / 100),
            Title=hit_args['title'],
            Description=hit_args['description'],
            Keywords=hit_args.get('keywords', ''))

        mturk.update_notification_settings(
            HITTypeId=hit_type_r['HITTypeId'],
            Notification={
                'Destination': current_app.config['MTURK_SNS_ARN'],
                'Transport': 'SNS',
                'Version': '2014-08-15',
                'EventTypes': [
                    'AssignmentAccepted',
                    'AssignmentAbandoned',
                    'AssignmentReturned',
                    'AssignmentSubmitted',
                    'AssignmentRejected',
                    'AssignmentApproved',
                    'HITCreated',
                    'HITExpired',
                    'HITReviewable',
                    'HITDisposed']},
            Active=True)

        task = Task(video=video, annotation_app=annotation_app, user=current_user, type_='HIT',
                    title=hit_args['title'], description=hit_args['description'],
                    keywords=hit_args.get('keywords', ''),
                    max_assignments=hit_args['max_assignments'],
                    reward=hit_args['reward'], args=hit_args.get('args', {}))
        db.session.add(task)
        db.session.commit()

        hit_response = mturk.create_hit_with_hit_type(
            HITTypeId=hit_type_r['HITTypeId'],
            MaxAssignments=hit_args['max_assignments'],
            LifetimeInSeconds=hit_args['lifetime_in_seconds'],
            Question=question_xml(app_url=annotation_app.url,
                                  query_string=urlencode({**{'videoId': video.id},
                                                          **hit_args.get('args', {})})))

        current_app.logger.info(f'Created HIT {hit_response["HIT"]["HITId"]} for Task {task.id}')

        task.hit_id = hit_response['HIT']['HITId']
        db.session.add(task)
        db.session.commit()

        return '', 201
    except Exception as e:
        db.session.rollback()
        raise


@task_bp.route('/task', methods=['GET'])
def list_tasks():
    query = sorted_query(Task)
    return paginated_response(query, lambda items: task_schema.dump(items, many=True))


@task_bp.route('/task/<uuid:task_id>/assignments', methods=['GET'])
def list_task_assignments(task_id):
    task = Task.query.get_or_404(task_id)
    query = sorted_query(Assignment, Assignment.query.filter_by(task=task))
    return paginated_response(query, lambda items: assignment_schema.dump(items, many=True))
